function [outputfh,output] = flute

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                 COMPUTATIONAL MODEL FOR THE FLUTE
%
% A simple computational flute model
% including a waveguide bore model and a static model for the air jet.
%
% Reflection function was implemented following Unto Laine's idea.
%
% Vesa V=E4lim=E4ki 1. - 2.1.1993, Last revision 30.8.1993
%
% Finger hole at an arbitrary point is included in this version !
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
starttime = clock;
%
% Constants:
%
fs = 44100.0;   % Sampling frequency (Hz)
T = 1/fs;       % Sampling interval (s)
%
L = 40;         % Length of the digital waveguide
Lx = 20;         % Distance from the mouth end to the finger hole (Lx real)
N = 38000       % Number of samples to be computed
%N = 10000 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Constant parameters of the Flute Model:
%
rb0 = -0.4794; rb1 = -0.2603;  % reflection filter coeffs.
% Second-order IIR Reflection Filter:
% rb=[-0.2260 -0.2339 0.01758];
% ra=[1 -0.6495 0.1271];
dca = 0.995;                  % DC killer (high-pass filter) coeff. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Parameters of the Fractional Delay 3-Port (Finger hole model):
%
Nx = 1;
%hi = [-0.1200 0.8400 0.2800];
%hi = [0 1 0];
Lx = 19;
% Interpolator parameters
hi = [0.25 0.75];    % Transitio toimii !!!!!! d=0.75
%hi = [0.5 0.5];   % Transitio ei toimi!       d=0.5
%hi = [1 0];         % Transitio toimii !!!!!! d=0
ar = -0.8632
%ar = 0.0
br = -(1+ar)
bri = 0
%Nbr1 = 5000;
Nbr1 = 12000;
Nbr2 = 24000;
%
wlast = 0;
i_line = 0;
d_line = 0;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Variable Parameters of the Flute Model:
%
loss_freq = 0.6;     % Coeff for the freq-dependent loss filter
loss_scale = 1 - loss_freq;  % Use this to scale DC gain of the loss filter
loss_gain = 0.94;    % freq-independent (DC) loss gain (0.92)
back_gain = -0.97;   % Real reflection coefficient for the mouth end
noise_gain = 4.0;    % Gain of the white noise signal
input_gain = 100.0;  % Maximum amplitude of input
voiced_gain = 0.02;  % Gain of the feedback from the tube to the nonlinearity
dir_fdback = -0.1;   % Gain from the nonlinearity back to itself
jet_scaler = 1.0;    % jet_length = jet_scaler * L
Ljet = round(jet_scaler * L); % Length of the jet line
integra = 0.0;       % Coefficient of the leaky integrator
                     % (when integra=0.0, inyegration is removed)
sigm_in = 0.001;     % Input gain of the sigmoid function
sigm_offset = -0.0;  % Offset of the sigmoid function
sigm_out = 300.0;    % Output gain of the sigmoid function
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Delay lines and pointers
%
upper = zeros(1,L+2);       % Upper trail of the digital waveguide
uptr = 1;                   % Pointer to the upper trail
lower = zeros(1,L+2);       % Lower trail of the digital waveguide
lptr = 1;                   % Pointer to the lower trail
jet_line = zeros(1,Ljet+2); % Delay line for the air jet
jptr = 1;                   % Pointer to the jet delay line
output = zeros(1,N);        % Output of the flute model
n = 1:N;                    % Discrete time indices for plotting
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Input of the flute model
%
pressure = noise_gain * 2*(rand(1,N) - 0.5);  % Input to the model
input_ampl = input_gain*[linspace(0,1,1) ones(1,N-1-2000) zeros(1,2000)];
%
%pressure = [-1000 zeros(1,N-1)];  % For measuring the impulse response
%input_gain = [1 zeros(1,N-1)];
%
%Pmax=2200; NR=500; Pd=Pmax/(NR-1); % linear rise + constant
%PP1=(0:Pd:Pmax); PP2=Pmax*ones(1,(N-NR)); P0=[PP1 PP2];
%
%Pmax=2200; NR=500; % raised cosine rise + constant
%xx=(0:(NR-1))/NR; PP1=Pmax*0.5*(1-cos(pi*xx));
%PP2=Pmax*ones(1,(N-NR)); P0=[PP1 PP2];
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% State variables
%
loss_delay = 0.0;    % Unit delay of the freq-dependent loss filter
refl_delayx = 0.0;   % First unit delay of the reflection function
refl_delayy = 0.0;   % Second unit delay of the reflection function
sigm_input = 0.0;    % Input of the sigmoid function
sigm_output = 0.0;   % Output of the sigmoid function
sigm_output1 = 0.0;  % Last output of the sigmoid function
DCX_output = 0.0;    % output of the DC killer filter
DCX_output1 = 0.0;   % Last output of the DC killer filter
output1 = 0.0;       % Last output of the model
integr_input = 0.0;  % Input of the leaky integrator
integr_delay = 0.0;  % Unit delay of the integrator
integr_output = 0.0; % Output of the integrator
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% The computation of the flute model is described below.
% The parts of the waveguide flute algorithm are as follows
% (in this order):
%
%  - add input (noise signal) and signal from the integrator
%  - multiply them by the input_gain
%  - delay the above result using jet delay line
%  - feed the delayed signal into the sigmoid waveshaper
%  - add the output of the sigmoid function and the reflecting signal
%    from the lower delay line of the waveguide
%  - FRACTIONAL DELAY 3-PORT HERE !!!
%  - update the pointers to the delay lines (circular buffers) and move
%    them to the other end of the lines
%  - take a sample from the end of the upper trail and feed it to
%    the reflection filter
%  - run the reflection filter one step
%  - subtract the output of the reflection function from its input
%  - this result is the output of the model which has to be differentiated
%  - feed the output of the reflection filter to the lower delay line
%  - take a sample out of the other end (i.e., beginning) of the lower line
%  - add it to the output of the sigmoid function (multiplied by dir_fdback)
%  - multiply this result by voiced_gain and feed it to the leaky integrator
%  - run one step of the integration (DC gain must be = 1)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Main loop
for i = 1:N;
   testi=0.01*i;
   if testi==round(testi); i,end;
   jet_line(jptr+1) = input_ampl(i) * (pressure(i) + integr_output);
   % Note: +1 needed with circular addressing in MATLAB only !
%   jptr = mod(jptr+Ljet,Ljet+1);  % jptr to the end of jet line
   jptr = rem(jptr+Ljet,Ljet+1);  % jptr to the end of jet line
   %
   % Sigmoid nonlinearity:
   sigm_input = jet_line(jptr+1);  % A sample from the jet line to the sigmoid
   sigm_output1 = sigm_output;     % Save previous sigm_output
   % Note: there is a bug in MATLAB's tanh function !
   tanh_input = input_ampl(i) * sigm_offset - sigm_in * sigm_input;
   if tanh_input > 10; tanh_input = 10; end;
   if tanh_input < -10; tanh_input = -10; end;
   sigm_output = sigm_out * tanh(tanh_input);
%   jptr = mod(jptr-Ljet-1,Ljet+1); % jptr back to the beginning & data move
   jptr = rem(jptr+Ljet-1,Ljet+1); % jptr back to the beginning & data move
   %
   % DC killer (a 1st order high-pass filter, direct form II)
   temp  = sigm_output + dca * DCX_output1;
   DCX_output = temp - DCX_output1;  % differentiation removes DC component
   DCX_output1 = temp;
   %
   % DCX_output = sigm_output; % Kokeilu, jossa DC-suodin poistettu
   %
   % Add (subtract) output of sigmoid function (DC killed) and
   %     reflected signal from lower delay line:
   upper(uptr+1) = DCX_output + back_gain * lower(lptr+1);
   %
   % FRACTIONAL DELAY 3-PORT:
   % i_line = delay line of the interpolator
   % hi = Lagrange interpolator coefficients
   % dei_line = delay line (state) of the deinterpolator
   % br & ar = coefficients of the reflection filter
   for m=1:Nx+1
     i_line(m) = upper(rem(uptr+Lx+m,L+1)+1) + lower(rem(lptr+Lx+1+m,L+1)+1);
   end;
   w1 = hi * i_line';
   if i == Nbr1,
     bri = br
   end; % Transition from bri=0 to bri=br
   if i == Nbr2,
     bri = 0.0
   end; % Transition from bri=br to bri=0
   w = bri*w1 - ar*wlast;
   wlast = w;
   outputfh(i) = w;
   dei_line = w * hi;
   for m=1:Nx+1
     duptr = rem(uptr+Lx+m,L+1);
     dlptr = rem(lptr+Lx+m,L+1);
     upper(duptr+1) = upper(duptr+1) + dei_line(m);
     lower(dlptr+1) = lower(dlptr+1) + dei_line(m);
   end;
   %
   % Pointers to the end of the waveguide & Data move:
   uptr = rem(uptr+L,L+1);   % uptr to the end of upper delay line
   lptr = rem(lptr+L+1,L+1); % lptr to the end of lower delay line & data move
   %
   % Boundary losses (1st order all-pole low-pass filter):
   loss_delay = loss_scale*loss_gain*upper(uptr+1) + loss_freq*loss_delay;
                        %%%loss_delay = loss_gain * upper(uptr+1); 
%TEST!!!!!!!
   %
   % Reflection filter (1st order IIR filter, direct form I implementation):
   refl_delayy = rb0 * loss_delay + rb1 * (refl_delayx - refl_delayy);
   refl_delayx = loss_delay;  % Update unit delay of the reflection filter
   %
   % Output of the model
   %  Subtract output of refl. filter from its input (and differentiate):
   output1 = loss_delay - refl_delayy;
   output(1,i) = output1;
   %
   % Reflected signal into the lower delay line:
   %lower(lptr) = refl_delayy;
   lower(lptr+1) = refl_delayy;
   %
   % Pointers to the beginning of the waveguide
   %uptr = rem(uptr-L-1,L+1); % uptr to the beg of upper delay line & data move
   uptr = rem(uptr+L-1,L+1); % uptr to the beg of upper delay line & data move
   %lptr = rem(lptr-L,L+1);   % lptr to the beg of lower delay line
   lptr = rem(lptr+L,L+1);   % lptr to the beg of lower delay line
   %
   % Feedback loop:
   integr_input = voiced_gain * (lower(lptr+1) + dir_fdback * DCX_output);
   %
   % Integration:
   integr_output = (1 - integra) * integr_input + integra * integr_output;
   %
end;  % end of the main loop
%
readytime = clock;
exetime = etime(readytime,starttime)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% plot(n,output);grid;
% title('Output of the Flute Model');
% xlabel('Time Index'); ylabel('Pressure');
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

